@@ -1,6 +1,7 @@
 /* Error handling for gpasm
    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005
    James Bowman, Craig Franklin
+   Copyright (C) 2012 Borut Razem
 
 This file is part of gputils.
 
@@ -33,22 +34,23 @@ struct error_list {
 
 static struct error_list *errorcodes_list = NULL;
 
-void add_code(int code)
+void
+add_code(int code)
 {
   struct error_list *new;
   struct error_list *list;
 
   if ((code <= -100) && (code >= -199)) {
-    gpwarning(GPW_DISABLE_ERROR, NULL);
+    gpvwarning(GPW_DISABLE_ERROR);
   } else {
     new = (struct error_list *)malloc(sizeof(*new));
     new->value = code;
     new->next  = NULL;
 
     if (errorcodes_list) {
-      /* the list has been started, scan the list for the end */ 
+      /* the list has been started, scan the list for the end */
       list = errorcodes_list;
-      while(list->next) {
+      while (list->next) {
         list = list->next;
       }
       list->next = new;  /* append the new value to the end of list */
@@ -58,7 +60,8 @@ void add_code(int code)
   }
 }
 
-static int check_code(int code)
+static int
+check_code(int code)
 {
   struct error_list *p;
   int print = 1;
@@ -66,11 +69,10 @@ static int check_code(int code)
   p = errorcodes_list;
 
   while(p) {
-
-    if (p->value == code){
-        print = 1;
+    if (p->value == code) {
+      print = 1;
     } else if (p->value == -(code)) {
-	print = 0;
+      print = 0;
     }
 
     p = p->next;
@@ -79,52 +81,98 @@ static int check_code(int code)
   return print;
 }
 
-char *gp_geterror(unsigned int code)
-{
+enum err_type_e {
+  et_error,
+  et_warning,
+  et_message
+};
 
+static void
+verr(enum err_type_e err_type, unsigned int code, const char *message, va_list ap)
+{
+  /* standard output */
+  if (!state.quiet) {
+    if (state.src)
+      printf("%s:%d:%s[%03d] %s", state.src->name, state.src->line_number,
+        (et_error == err_type) ? "Error" : (et_warning == err_type) ? "Warning" : "Message",
+        code, ((et_error == err_type) ? "  " : ""));
+    else
+      printf("%s[%03d] %s",
+        (et_error == err_type) ? "Error" : (et_warning == err_type) ? "Warning" : "Message",
+        code, ((et_error == err_type) ? "  " : ""));
+    vprintf(message, ap);
+    putchar('\n');
+  }
+}
+
+static void
+err(enum err_type_e err_type, unsigned int code, const char *message)
+{
+  /* standard output */
+  if (!state.quiet) {
+    if (state.src)
+      printf("%s:%d:%s[%03d] %s%s\n", state.src->name, state.src->line_number,
+        (et_error == err_type) ? "Error" : (et_warning == err_type) ? "Warning" : "Message",
+        code, ((et_error == err_type) ? "  " : ""), message);
+    else
+      printf("%s[%03d] %s%s\n",
+        (et_error == err_type) ? "Error" : (et_warning == err_type) ? "Warning" : "Message",
+        code, ((et_error == err_type) ? "  " : ""), message);
+  }
+}
+
+static char *
+gp_geterror(unsigned int code)
+{
   switch(code) {
+  case GPE_USER:
+    return "ERROR: (%s)";
+  case GPE_NOENT:
+    return "Cannot open file (Include File \"%s\" not found)";
+  case GPE_STRCPLX:
+    return "String substitution too complex";
+  case GPE_BADCHAR:
+    return "Illegal character (%c).";
   case GPE_OPENPAR:
     return "Unmatched (";
   case GPE_CLOSEPAR:
     return "Unmatched )";
-  case GPE_BADCHAR:
-    return "Illegal character.";
   case GPE_NOSYM:
-    return "Symbol not previously defined.";
+    return "Symbol not previously defined (%s)";
   case GPE_DIVBY0:
     return "Divide by zero.";
   case GPE_DUPLAB:
-    return "Duplicate label or redefining symbol that cannot be redefined.";
+    return "Duplicate label (\"%s\" or redefining symbol that cannot be redefined)";
   case GPE_DIFFLAB:
-    return "Address label duplicated or different in second pass.";
+    return "Address label duplicated or different in second pass (%s)";
   case GPE_ADDROVF:
-    return "Address wrapped around 0. ";
+    return "Address wrapped around 0.";
   case GPE_ADDROVR:
-    return "Overwriting previous address contents.";
+    return "Overwriting previous address contents (%04X)";
   case GPE_BAD_CALL_ADDR:
     return "Call or jump not allowed at this address (must be in low half of page)";
   case GPE_ILLEGAL_LABEL:
-    return "Illegal label.";
+    return "Illegal label (%s)";
   case GPE_ILLEGAL_DIR:
-    return "Illegal directive (Not Valid for this processor).";
+    return "Illegal directive (%s)";
   case GPE_ILLEGAL_ARGU:
-    return "Illegal argument.";
+    return "Illegal argument (%s)";
   case GPE_ILLEGAL_COND:
     return "Illegal condition.";
   case GPE_RANGE:
-    return "Argument out of range.";
+    return "Argument out of range";
   case GPE_TOO_MANY_ARGU:
-    return "Too many arguments.";
+    return "Too many arguments";
   case GPE_MISSING_ARGU:
-    return "Missing argument(s).";
+    return "Missing argument(s)";
   case GPE_EXPECTED:
     return "Expected";
   case GPE_EXTRA_PROC:
     return "Processor type previously defined.";
   case GPE_UNDEF_PROC:
-    return "Processor type is undefined.";
+    return "Processor type is undefined";
   case GPE_UNKNOWN_PROC:
-    return "Unknown processor.";
+    return "Unknown processor (%s)";
   case GPE_IHEX:
     return "Hex file format INHX32 required.";
   case GPE_NO_MACRO_NAME:
@@ -139,12 +187,14 @@ char *gp_geterror(unsigned int code)
     return "Unmatched ENDM.";
   case GPE_OBJECT_ONLY:
     return "Directive only allowed when generating an object file.";
+  case GPE_LABEL_IN_SECTION:
+    return"Labels must be defined in a code or data section when making an object file";
   case GPE_UNRESOLVABLE:
-    return "Operand contains unresolvable labels or is too complex.";
+    return "Operand contains unresolvable labels or is too complex";
   case GPE_WRONG_SECTION:
-    return "Executable code and data must be defined in an appropriate section.";
+    return "Executable code and data must be defined in an appropriate section";
   case GPE_CONTIG_SECTION:
-    return "Each object file section must be contiguous.";
+    return "Each object file section must be contiguous (section %s)";
   case GPE_MUST_BE_LABEL:
     return "Operand must be an address label.";
   case GPE_ORG_ODD:
@@ -155,78 +205,92 @@ char *gp_geterror(unsigned int code)
     return "__CONFIG directives must be contiguous.";
   case GPE_CONTIG_IDLOC:
     return "__IDLOC directives must be contiguous.";
+  case GPE_NO_EXTENDED_MODE:
+    return "Extended mode not available for this device.";
   case GPE_MISSING_BRACKET:
-    return "Square brackets required around offset operand.";
+    return "Square brackets required around offset operand";
+  case GPE_CONSTANT:
+    return "Expression within brackets must be constant";
+  case GPE_IDLOCS_ORDER:
+    return "__IDLOCS directives must be listed in ascending order.";
   case GPE_CONFIG_UNKNOWN:
     return "An error with the CONFIG directive occured.";
   case GPE_CONFIG_usCONFIG:
     return "You cannot mix CONFIG and __CONFIG directives.";
+  case GPE_RES_ODD_PIC16EA:
+    return "RES directive cannot reserve odd number of bytes in PIC18 absolute mode.";
   case GPE_UNKNOWN:
+    return "";
+
+  /* gputils special errors */
+  case GPE_INTERNAL:
+    return "Internal error: %s";
+  case GPE_PARSER:
+    return "Parser error: %s";
+  case GPE_SCANNER:
+    return "Scanner error: %s";
+
   default:
     return "UNKNOWN";
   }
 }
 
-void gperror(unsigned int code,
-	     char *message)
+void
+gperror(unsigned int code, char *message)
 {
-  char full_message[BUFSIZ];
-
   if (state.pass == 2) {
-    if(message == NULL)
+    if (message == NULL)
       message = gp_geterror(code);
 
-#ifndef GP_USER_ERROR
     /* standard output */
-    if (!state.quiet) {
-      if (state.src)
-        snprintf(full_message, sizeof(full_message),
-                 "%s:%d:Error [%03d] %s",
-                 state.src->name,
-                 state.src->line_number,
-                 code,
-                 message);
-      else
-        snprintf(full_message, sizeof(full_message),
-                 "Error [%03d] %s",
-                 code,
-                 message);
-
-      printf("%s\n", full_message);
-    }
-#else
-    user_error(code, message);
-#endif
+    err(et_error, code, message);
 
     /* list file output */
-    snprintf(full_message, sizeof(full_message),
-	     "Error [%03d] : %s",
-	     code,
-	     message);
-
-    lst_line(full_message);
+    lst_line("Error[%03d]  : %s", code, message);
 
     state.num.errors++;
   }
 }
 
-char *gp_getwarning(unsigned int code)
+void
+gpverror(unsigned int code, ...)
+{
+  if (state.pass == 2) {
+    va_list(ap);
+    const char *message = gp_geterror(code);
+
+    /* standard output */
+    va_start(ap, code);
+    verr(et_error, code, message, ap);
+    va_end(ap);
+
+    /* list file output */
+    va_start(ap, code);
+    lst_err_line("Error", code, message, ap);
+    va_end(ap);
+
+    state.num.errors++;
+  }
+}
+
+static char *
+gp_getwarning(unsigned int code)
 {
   switch(code) {
   case GPW_NOT_DEFINED:
-    return "Symbol not previously defined.";
+    return "Symbol not previously defined (%s)";
   case GPW_RANGE:
-    return "Argument out of range. Least significant bits used.";
+    return "Argument out of range.  Least significant bits used.";
   case GPW_OP_COLUMN_ONE:
-    return "Found opcode in column 1.";
+    return "Found opcode in column 1. (%s)";
   case GPW_DIR_COLUMN_ONE:
-    return "Found directive in column 1.";
+    return "Found directive in column 1. (%s)";
   case GPW_MACRO_COLUMN_ONE:
-    return "Found call to macro in column 1.";
+    return "Found call to macro in column 1. (%s)";
   case GPW_LABEL_COLUMN:
     return "Found label after column 1.";
   case GPW_MISSING_QUOTE:
-    return "Missing quote.";
+    return "Missing quote";
   case GPW_EXTRANEOUS:
     return "Extraneous arguments on the line.";
   case GPW_EXPECTED:
@@ -249,54 +313,33 @@ char *gp_getwarning(unsigned int code)
     return "Redefining processor.";
   case GPW_NOT_RECOMMENDED:
     return "Use of this instruction is not recommended.";
+  case GPW_WORD_ALIGNED:
+    return "Destination address must be word aligned.";
   case GPW_INVALID_ROM:
     return "Invalid ROM location specified.";
-  case GPW_UNKNOWN:
+
+  /* gputils special warnings */
+  case GPW_BANK_PAGE_SEL_AFTER_SKIP:
+    return "%s after skip instruction. I this really what you intended?";
+
   default:
     return "UNKNOWN";
   }
 }
 
-void gpwarning(unsigned int code,
-	       char *message)
+void
+gpwarning(unsigned int code, char *message)
 {
-  char full_message[BUFSIZ];
-
-  if (state.pass ==2) {
-
+  if (state.pass == 2) {
     if ((state.error_level <= 1) && check_code(code)) {
-      if(message == NULL)
+      if (message == NULL)
         message = gp_getwarning(code);
 
-#ifndef GP_USER_WARNING
       /* standard output */
-      if (!state.quiet) {
-        if (state.src)
-          snprintf(full_message, sizeof(full_message),
-                   "%s:%d:Warning [%03d] %s",
-                   state.src->name,
-                   state.src->line_number,
-                   code,
-                   message);
-        else
-          snprintf(full_message, sizeof(full_message),
-                   "Warning [%03d] %s",
-                   code,
-                   message);
-
-        printf("%s\n", full_message);
-      } 
-#else
-      user_warning(code, message);
-#endif
+      err(et_warning, code, message);
 
       /* list file output */
-      snprintf(full_message, sizeof(full_message),
-	       "Warning [%03d] : %s",
-	       code,
-	       message);
-
-      lst_line(full_message);
+      lst_line("Warning[%03d]: %s", code, message);
 
       state.num.warnings++;
     } else {
@@ -305,17 +348,42 @@ void gpwarning(unsigned int code,
   }
 }
 
-char *gp_getmessage(unsigned int code)
+void
+gpvwarning(unsigned int code, ...)
+{
+  if (state.pass == 2) {
+    if ((state.error_level <= 1) && check_code(code)) {
+      va_list(ap);
+      const char *message = gp_getwarning(code);
+
+      /* standard output */
+      va_start(ap, code);
+      verr(et_warning, code, message, ap);
+      va_end(ap);
+
+      va_start(ap, code);
+      lst_err_line("Warning", code, message, ap);
+      va_end(ap);
+
+      state.num.warnings++;
+    } else {
+      state.num.warnings_suppressed++;
+    }
+  }
+}
+
+static char *
+gp_getmessage(unsigned int code)
 {
   switch(code) {
   case GPM_USER:
-    return "MESSAGE:";
+    return "MESSAGE: (%s)";
   case GPM_BANK:
-    return "Register in operand not in bank 0. Ensure bank bits are correct.";
+    return "Register in operand not in bank 0.  Ensure that bank bits are correct.";
   case GPM_RANGE:
-    return "Program word too large. Truncated to core size.";
+    return "Program word too large.  Truncated to core size. (%04X)";
   case GPM_IDLOC:
-    return "ID Locations value too large. Last four hex digits used.";
+    return "ID Locations value too large. Last four hex digits used. (%04X)";
   case GPM_NOF:
     return "Using default destination of 1 (file).";
   case GPM_PAGE:
@@ -329,59 +397,58 @@ char *gp_getmessage(unsigned int code)
   case GPM_SUPRAM:
     return "Superceding current maximum RAM and RAM map.";
   case GPM_EXTPAGE:
-    return "Page or Bank selection not needed for this device.";
+    return "Page or Bank selection not needed for this device.  No code generated.";
   case GPM_CBLOCK:
     return "CBLOCK constants will start with a value of 0.";
   case GPM_W_MODIFIED:
     return "W Register modified.";
   case GPM_SPECIAL_MNEMONIC:
-    return "Special Instruction Mnemonic used."; 
+    return "Special Instruction Mnemonic used.";
   case GPM_UNKNOWN:
   default:
     return "UNKNOWN MESSAGE";
   }
 }
 
-void gpmessage(unsigned int code,
-	       char *message)
+void
+gpmessage(unsigned int code, char *message)
 {
-  char full_message[BUFSIZ];
-
-  if (state.pass==2) {
-
+  if (state.pass == 2) {
     if ((state.error_level == 0) && check_code(code)){
-      if(message == NULL)
+      if (message == NULL)
         message = gp_getmessage(code);
 
-#ifndef GP_USER_MESSAGE
       /* standard output */
-      if (!state.quiet) {
-        if (state.src)
-          snprintf(full_message, sizeof(full_message),
-                   "%s:%d:Message [%03d] %s",
-                   state.src->name,
-                   state.src->line_number,
-                   code,
-                   message);
-        else
-          snprintf(full_message, sizeof(full_message),
-                   "Message [%03d] %s",
-                   code,
-                   message);
-
-        printf("%s\n", full_message);
-      }
-#else
-      user_message(code, message);
-#endif
+      err(et_message, code, message);
 
       /* list file output */
-      snprintf(full_message, sizeof(full_message),
-               "Message [%03d] : %s",
-               code,
-               message);
+      lst_line("Message[%03d]: %s", code, message);
 
-      lst_line(full_message);
+      state.num.messages++;
+    } else {
+      state.num.messages_suppressed++;
+    }
+  }
+}
+
+void
+gpvmessage(unsigned int code, ...)
+{
+  if (state.pass == 2) {
+
+    if ((state.error_level == 0) && check_code(code)){
+      va_list(ap);
+      const char *message = gp_getmessage(code);
+
+      /* standard output */
+      va_start(ap, code);
+      verr(et_message, code, message, ap);
+      va_end(ap);
+
+      /* list file output */
+      va_start(ap, code);
+      lst_err_line("Message", code, message, ap);
+      va_end(ap);
 
       state.num.messages++;
     } else {
