@@ -27,6 +27,7 @@ Boston, MA 02111-1307, USA.  */
 #include "scan.h"
 #include "lst.h"
 #include "map.h"
+#include "script.h"
 
 struct gplink_state state;
 static int processor_mismatch_warning = 1;
@@ -35,12 +36,14 @@ int yyparse(void);
 extern int yydebug;
 
 /* return the number of missing symbols */
-int count_missing(void)
+int
+count_missing(void)
 {
   return state.symbol.missing->count;
 }
 
-void object_append(gp_object_type *file, char *name)
+void
+object_append(gp_object_type *file, char *name)
 {
   /* append the entry to the list */
   if (state.object == NULL) {
@@ -55,33 +58,31 @@ void object_append(gp_object_type *file, char *name)
       list = list->next;
     }
     list->next = file;
-    
+
     if (file->class != state.class) {
       gp_error("processor family mismatch in \"%s\"", file->filename);
     } else if ((0 != processor_mismatch_warning) &&
                (file->processor != state.processor)) {
       gp_warning("processor mismatch in \"%s\"", file->filename);
     }
-
   }
 
   if (state.optimize.weak_symbols) {
     gp_coffopt_remove_weak(file);
   }
-
-  return;
 }
 
-void archive_append(gp_archive_type *file, char *name)
+void
+archive_append(gp_archive_type *file, char *name)
 {
   struct archivelist *new;
 
-  /* make the new entry */  
+  /* make the new entry */
   new = (struct archivelist *)malloc(sizeof(*new));
-  new->name = strdup(name); 
+  new->name = strdup(name);
   new->archive = file;
   new->next = NULL;
-  
+
   /* append the entry to the list */
   if (state.archives == NULL) {
     state.archives = new;
@@ -93,13 +94,11 @@ void archive_append(gp_archive_type *file, char *name)
     }
     list->next = new;
   }
-
-  return;
 }
 
 /* Scan the archive for missing symbol definitions.  This has to be done
    recursively.  The order of the archive members is unknown and there
-   might be inter member dependancies.  Scan the archive muliple times.  
+   might be inter member dependancies.  Scan the archive muliple times.
    Stop whenever a complete pass through the archive happens and no
    objects are added. */
 
@@ -112,7 +111,7 @@ scan_index(struct symbol_table *table, gp_archive_type *archive)
   gp_object_type *object;
   int i;
   int num_added = 1; /* initalize to 1 so while loop can be entered */
-  char *name;
+  const char *name;
   char *object_name;
   gp_boolean modified = false;
 
@@ -120,35 +119,35 @@ scan_index(struct symbol_table *table, gp_archive_type *archive)
     num_added = 0;
     for (i = 0; i < HASH_SIZE; i++) {
       for (s = state.symbol.missing->hash_table[i]; s; s = s->next) {
-	name = get_symbol_name(s);
+        name = get_symbol_name(s);
         assert(name != NULL);
         /* Search for missing symbol name in archive symbol table */
         m = get_symbol(table, name);
         if (m != NULL) {
-          /* Fetch the archive member, convert its binary data to an object 
-	     file, and add the object to the object list */ 
+          /* Fetch the archive member, convert its binary data to an object
+             file, and add the object to the object list */
           member = get_symbol_annotation(m);
           object_name = gp_archive_member_name(member);
-          object = gp_convert_file(object_name, member->file);
-          object_append(object, object_name);	
-          gp_link_add_symbols(state.symbol.definition, 
-                              state.symbol.missing, 
-			      object);		
+          object = gp_convert_file(object_name, &member->data);
+          object_append(object, object_name);
+          gp_link_add_symbols(state.symbol.definition,
+                              state.symbol.missing,
+                              object);
           /* The symbol tables have been modified.  Need to take another
-	     pass to make sure we get everything. */
+             pass to make sure we get everything. */
           num_added++;
           modified = true;
           free(object_name);
           /* This branch of the table has been modified. Go to the next one */
-	  break;
-	}      
+          break;
+        }
       }
     }
   }
 
   return modified;
 }
-    
+
 gp_boolean
 scan_archive(gp_archive_type *archive, char *name)
 {
@@ -159,7 +158,7 @@ scan_archive(gp_archive_type *archive, char *name)
   /* If necessary, build a symbol index for the archive. */
   if (gp_archive_have_index(archive) == 0) {
     struct symbol_table *archive_tbl = NULL;
-  
+
     archive_tbl = push_symbol_table(NULL, true);
     gp_archive_make_index(archive, archive_tbl);
     archive = gp_archive_add_index(archive_tbl, archive);
@@ -190,8 +189,6 @@ remove_linker_symbol(char *name)
   if (sym != NULL) {
     gp_link_remove_symbol(state.symbol.missing, name);
   }
-
-  return;
 }
 
 /* Add a symbol the linker created to the symbol table. */
@@ -214,8 +211,6 @@ add_linker_symbol(char *name)
 
   assert(found != NULL);
   gp_link_add_symbol(state.symbol.definition, found, NULL);
-
-  return;
 }
 
 /* Search the object list for an idata section. */
@@ -231,34 +226,33 @@ search_idata(void)
     while (section != NULL) {
       if (section->flags & STYP_DATA) {
         state.has_idata = true;
-        return;    
+        return;
       }
       section = section->next;
     }
     list = list->next;
   }
-
-  return;
 }
 
-/* Build the symbol tables.  Determine which objects from the archives are 
+/* Build the symbol tables.  Determine which objects from the archives are
    required for linking */
-    
-void build_tables(void)
+
+void
+build_tables(void)
 {
   gp_object_type *list = state.object;
   struct archivelist *arlist;
   gp_boolean modified;
   int i;
   struct symbol *s;
-  char *name;
+  const char *name;
   gp_coffsymbol_type *var;
 
-  /* Create the object file symbol tables */  
+  /* Create the object file symbol tables */
   while (list != NULL) {
-    gp_link_add_symbols(state.symbol.definition, 
-                        state.symbol.missing, 
-			list);
+    gp_link_add_symbols(state.symbol.definition,
+                        state.symbol.missing,
+                        list);
     list = list->next;
   }
 
@@ -317,39 +311,37 @@ void build_tables(void)
     }
     exit(1);
   }
-
-  return;
 }
 
 /* Read a coff object or archive.  gplink doesn't care about file extensions.
-   This allows alternate extensions such as .a archives and .obj coff 
+   This allows alternate extensions such as .a archives and .obj coff
    objects.  */
 
-void gplink_open_coff(char *name)
+void
+gplink_open_coff(const char *name)
 {
   gp_object_type *object;
   gp_archive_type *archive;
   FILE *coff;
-  char file_name[BUFSIZ];
-  
+  char file_name[PATH_MAX + 1];
+
   strncpy(file_name, name, sizeof(file_name));
-  
+
   coff = fopen(file_name, "rb");
-  if ((coff == NULL) && (strchr(file_name, PATH_CHAR) == 0)) { 
+  if ((coff == NULL) && (strchr(file_name, PATH_CHAR) == 0)) {
     /* If no "/" in name, try searching include pathes */
     int i;
 
     for(i = 0; i < state.numpaths; i++) {
-      strncpy(file_name, state.paths[i], sizeof(file_name));
-      strncat(file_name, COPY_CHAR, sizeof(file_name));
-      strncat(file_name, name, sizeof(file_name));
+      snprintf(file_name, sizeof(file_name),
+               "%s" COPY_CHAR "%s", state.paths[i], name);
       coff = fopen(file_name, "rb");
       if (coff != NULL) {
         break;
       }
-    }  
+    }
   }
-  
+
   if (coff == NULL) {
     perror(name);
     exit(1);
@@ -360,31 +352,29 @@ void gplink_open_coff(char *name)
   switch(gp_identify_coff_file(file_name)) {
   case object_file_v2:
   case object_file:
-    /* read the object */  
+    /* read the object */
     object = gp_read_coff(file_name);
     object_append(object, file_name);
     break;
   case archive_file:
-    /* read the archive */  
+    /* read the archive */
     archive = gp_archive_read(file_name);
     archive_append(archive, file_name);
     break;
   case sys_err_file:
-    gp_error("can't open file \"%s\"", file_name);     
+    gp_error("can't open file \"%s\"", file_name);
     break;
   case unknown_file:
-    gp_error("\"%s\" is not a valid coff object or archive", file_name);     
-    break; 
+    gp_error("\"%s\" is not a valid coff object or archive", file_name);
+    break;
   default:
     assert(0);
   }
-
-}   
+}
 
 static void
 set_optimize_level(void)
 {
-
   /* default */
   state.optimize.dead_sections = false;
   state.optimize.weak_symbols = false;
@@ -403,45 +393,47 @@ set_optimize_level(void)
   default:
     gp_error("invalid optimization level");
   }
-
-  return;
 }
 
-#define GET_OPTIONS "?a:cdf:hI:lmo:O:qrs:t:vw"
+#define GET_OPTIONS "?a:cdf:hI:lmo:O:qrs:t:u:vw"
 
-  static struct option longopts[] =
-  {
-    { "hex-format",          1, 0, 'a' },
-    { "object",              0, 0, 'c' },
-    { "debug",               0, 0, 'd' },
-    { "fill",                1, 0, 'f' },
-    { "help",                0, 0, 'h' },
-    { "include",             1, 0, 'I' },
-    { "no-list",             0, 0, 'l' },
-    { "map",                 0, 0, 'm' },
-    { "output",              1, 0, 'o' },
-    { "optimize",            1, 0, 'O' },
-    { "quiet",               0, 0, 'q' },
-    { "use-shared",          0, 0, 'r' },
-    { "script",              1, 0, 's' },
-    { "stack",               1, 0, 't' },
-    { "version",             0, 0, 'v' },
-    { "processor-mismatch",  0, 0, 'w' },
-    { 0, 0, 0, 0 }
-  };
+enum {
+  OPT_MPLINK_COMPATIBLE = 0x100
+};
+
+static struct option longopts[] =
+{
+  { "hex-format",          required_argument, NULL, 'a' },
+  { "object",              no_argument,       NULL, 'c' },
+  { "debug",               no_argument,       NULL, 'd' },
+  { "fill",                required_argument, NULL, 'f' },
+  { "help",                no_argument,       NULL, 'h' },
+  { "include",             required_argument, NULL, 'I' },
+  { "no-list",             no_argument,       NULL, 'l' },
+  { "map",                 no_argument,       NULL, 'm' },
+  { "output",              required_argument, NULL, 'o' },
+  { "optimize",            required_argument, NULL, 'O' },
+  { "quiet",               no_argument,       NULL, 'q' },
+  { "use-shared",          no_argument,       NULL, 'r' },
+  { "script",              required_argument, NULL, 's' },
+  { "stack",               required_argument, NULL, 't' },
+  { "macro",               required_argument, NULL, 'u' },
+  { "version",             no_argument,       NULL, 'v' },
+  { "processor-mismatch",  no_argument,       NULL, 'w' },
+  { "mplink-compatible",   no_argument,       NULL, OPT_MPLINK_COMPATIBLE },
+  { NULL,                  no_argument,       NULL, '\0'}
+};
 
 void
 init(void)
 {
-
   gp_init();
 
   /* initialize */
   gp_date_string(state.startdate, sizeof(state.startdate));
   state.hex_format = inhx32;
   state.numpaths = 0;
-  state.byte_addr = 0;
-  state.processor = no_processor;
+  state.processor = NULL;
   state.optimize.level = 1;
   state.codfile = normal;
   state.hexfile = normal;
@@ -453,32 +445,50 @@ init(void)
   state.has_stack = false;
   state.stack_size = 0;
   state.has_idata = false;
-  state.srcfilename = NULL;
+  state.srcfilenames = NULL;
   state.object  = NULL;
   state.archives = NULL;
 
   /* set default output filename to be a.o, a.hex, a.cod, a.map */
   strncpy(state.basefilename, "a", sizeof(state.basefilename));
 
+  state.ifdef = NULL;
+  state.script_symbols = push_symbol_table(NULL, false);
+
   /* The symbols are case sensitive */
   state.symbol.definition = push_symbol_table(NULL, false);
   state.symbol.missing = push_symbol_table(NULL, false);
   state.section.definition = push_symbol_table(NULL, false);
   state.section.logical = push_symbol_table(NULL, false);
-
-  return;
 }
 
-void gplink_add_path(char *path)
+void
+gplink_add_path(const char *path)
 {
-  if(state.numpaths < MAX_PATHS) {
+  if (state.numpaths < MAX_PATHS) {
     state.paths[state.numpaths++] = strdup(path);
   } else {
     gp_error("too many -I paths");
   }
 }
 
-void show_usage(void)
+/* I have this take the func in anticipation of the option
+  printf("  -z, --defsym symbol=value      Add symbol value to symbol table.\n");
+*/
+static void
+parse_define(const char *optarg, void (*func)(const char* name, long value))
+{
+  long value = 0;
+  char *pc = strchr(optarg, '=');
+  if (pc) {
+    *pc++ = 0;
+    value = strtol(pc, &pc, 10);
+  }
+  func(optarg, value);
+}
+
+void
+show_usage(void)
 {
   printf("Usage: gplink [options] [objects] [libraries] \n");
   printf("Options: [defaults in brackets after descriptions]\n");
@@ -490,12 +500,14 @@ void show_usage(void)
   printf("  -I DIR, --include DIR          Specify include directory.\n");
   printf("  -l, --no-list                  Disable list file output.\n");
   printf("  -m, --map                      Output a map file.\n");
+  printf("      --mplink-compatible         MPLINK compatibility mode\n");
   printf("  -o FILE, --output FILE         Alternate name of output file.\n");
   printf("  -O OPT, --optimize OPT         Optimization level [1].\n");
   printf("  -q, --quiet                    Quiet.\n");
   printf("  -r, --use-shared               Use shared memory if necessary.\n");
   printf("  -s FILE, --script FILE         Linker script.\n");
   printf("  -t SIZE, --stack SIZE          Create a stack section.\n");
+  printf("  -u, --macro symbol[=value]     Add macro value for script.\n");
   printf("  -v, --version                  Show version.\n");
   printf("  -w, --processor-mismatch       Disable \"processor mismatch\" warning.\n");
   printf("\n");
@@ -505,12 +517,12 @@ void show_usage(void)
   } else {
     printf("Default linker script path NOT SET\n");
   }
-  if (gp_lkr_path) {
+  if (gp_lib_path) {
     printf("Default library path %s\n", gp_lib_path);
   } else {
     printf("Default library path NOT SET\n");
   }
-  printf("\n");    
+  printf("\n");
 #endif
   printf("Report bugs to:\n");
   printf("%s\n", PACKAGE_BUGREPORT);
@@ -555,13 +567,16 @@ process_args( int argc, char *argv[])
                  optarg);
       }
       break;
+
     case 'c':
       state.objfile = normal;
       break;
+
     case 'd':
       gp_debug_disable = false;
       yydebug = 1;
       break;
+
     case 'f':
       state.fill_value = strtol(optarg, &pc, 16);
       if ((pc == NULL) || (*pc != '\0')) {
@@ -572,37 +587,63 @@ process_args( int argc, char *argv[])
         state.fill_enable = true;
       }
       break;
+
     case '?':
     case 'h':
       usage = true;
       break;
+
     case 'I':
       gplink_add_path(optarg);
       break;
+
     case 'l':
       state.lstfile = suppress;
       break;
+
     case 'm':
       state.mapfile = normal;
       break;
+
     case 'o':
       strncpy(state.basefilename, optarg, sizeof(state.basefilename));
       pc = strrchr(state.basefilename, '.');
       if (pc)
         *pc = 0;
       break;
+
     case 'O':
       /* do nothing */
       break;
+
     case 'q':
       gp_quiet = true;
       break;
+
     case 'r':
       gp_relocate_to_shared = true;
       break;
+
     case 's':
-      state.srcfilename = optarg;
+      {
+        struct srcfns *fn;
+
+        fn = malloc(sizeof (struct srcfns));
+        fn->filename = strdup(optarg);
+        fn->next = NULL;
+
+        if (NULL == state.srcfilenames)
+          state.srcfilenames = fn;
+        else {
+          struct srcfns *p;
+
+          for (p = state.srcfilenames; p->next; p = p->next)
+            ;
+          p->next = fn;
+        }
+      }
       break;
+
     case 't':
       state.stack_size = strtol(optarg, &pc, 10);
       if ((pc == NULL) || (*pc != '\0')) {
@@ -611,13 +652,23 @@ process_args( int argc, char *argv[])
         state.has_stack = true;
       }
       break;
+
+    case 'u':
+      parse_define(optarg, add_script_macro);
+      break;
+
     case 'v':
       fprintf(stderr, "%s\n", GPLINK_VERSION_STRING);
       exit(0);
       break;
+
     case 'w':
       processor_mismatch_warning = 0;
       break;
+
+    case OPT_MPLINK_COMPATIBLE:
+      state.mplink_compatible = true;
+      break;
     }
     if (usage)
       break;
@@ -626,18 +677,26 @@ process_args( int argc, char *argv[])
   /* check if the first file is the linker script */
   if (argv[optind]) {
     pc = strrchr(argv[optind], '.');
-    if (strcasecmp(pc, ".lkr") == 0) {
-      if (state.srcfilename == NULL) {
-        state.srcfilename = argv[optind++];
-      } else {
-        gp_error("linker script specified twice (%s and %s)",
-                 state.srcfilename,
-                 argv[optind]);
+    if (pc && strcasecmp(pc, ".lkr") == 0) {
+      struct srcfns *fn;
+
+      fn = malloc(sizeof (struct srcfns));
+      fn->filename = strdup(argv[optind++]);
+      fn->next = NULL;
+
+      if (NULL == state.srcfilenames)
+        state.srcfilenames = fn;
+      else {
+        struct srcfns *p;
+
+        for (p = state.srcfilenames; p->next; p = p->next)
+          ;
+        p->next = fn;
       }
     }
   }
 
-  if ((state.srcfilename == NULL) &&
+  if ((state.srcfilenames == NULL) &&
       (optind >= argc)) {
     /* No linker script was specified and no object filenames were provided,
        so print the usage */
@@ -658,75 +717,72 @@ process_args( int argc, char *argv[])
     gplink_add_path(gp_lkr_path);
   }
 
-  /* Open all objects and archives in the file list. */ 
+  /* Open all objects and archives in the file list. */
   for ( ; optind < argc; optind++) {
     gplink_open_coff(argv[optind]);
   }
-
 }
 
-int
+gp_boolean
 linker(void)
 {
   MemBlock *data, *program;
 
   /* setup output filenames */
-  strncpy(state.hexfilename, state.basefilename, sizeof(state.hexfilename));
-  strncat(state.hexfilename, ".hex", sizeof(state.hexfilename));
-  strncpy(state.mapfilename, state.basefilename, sizeof(state.mapfilename));
-  strncat(state.mapfilename, ".map", sizeof(state.mapfilename));
-  strncpy(state.objfilename, state.basefilename, sizeof(state.objfilename));
-  strncat(state.objfilename, ".cof", sizeof(state.objfilename));
+  snprintf(state.hexfilename, sizeof(state.hexfilename),
+           "%s.hex", state.basefilename);
+  snprintf(state.mapfilename, sizeof(state.mapfilename),
+           "%s.map", state.basefilename);
+  snprintf(state.objfilename, sizeof(state.objfilename),
+           "%s.cof", state.basefilename);
 
   /* Read the script */
-  if (state.srcfilename) {
-    open_src(state.srcfilename, 0);
-    yyparse();
+  if (state.srcfilenames) {
+    struct srcfns *p = state.srcfilenames;
+
+    do {
+      open_src(p->filename, 0);
+      yyparse();
+      p = p->next;
+    } while (NULL != p);
 #ifdef USE_DEFAULT_PATHS
   } else if ((state.object) && (gp_lkr_path)) {
     /* The processor is known because an object was on the command line. So
        use one of the default scripts that are distributed with gputils. */
     char file_name[BUFSIZ];
-    char *script_name;
-    
-    assert(state.processor != no_processor); 
+    const char *script_name;
+
+    assert(state.processor);
     script_name = gp_processor_script(state.processor);
     if (script_name == NULL) {
       gp_error("linker script not specified and can't determine default script");
-      return EXIT_FAILURE; 
+      return EXIT_FAILURE;
     }
-    strncpy(file_name, gp_lkr_path, sizeof(file_name));
-    strncat(file_name, COPY_CHAR, sizeof(file_name));
-    strncat(file_name, script_name, sizeof(file_name));
+    snprintf(file_name, sizeof(file_name),
+             "%s" COPY_CHAR "%s", gp_lkr_path, script_name);
     gp_message("using default linker script \"%s\"", file_name);
     open_src(file_name, 0);
     yyparse();
 #endif
   } else {
     /* The user must supply the linker script name.  The processor isn't
-       commanded so the linker has no way to pick. */ 
+       commanded so the linker has no way to pick. */
     gp_error("linker script not specified");
-    return EXIT_FAILURE; 
+    return EXIT_FAILURE;
   }
 
   if (state.object == NULL) {
     gp_error("missing input object file");
-    return EXIT_FAILURE; 
-  }
-
-  if (state.class == PROC_CLASS_PIC16E) {
-    state.byte_addr = 1;
-  } else {
-    state.byte_addr = 0;
+    return EXIT_FAILURE;
   }
 
   /* An error occured while reading the input files, no need to continue */
   if (gp_num_errors)
     return EXIT_FAILURE;
 
-  /* Construct the symbol tables. Determine which archive members are 
+  /* Construct the symbol tables. Determine which archive members are
      required to resolve external references. */
-  build_tables();  
+  build_tables();
 
   /* combine all object files into one object */
   gp_cofflink_combine_objects(state.object);
@@ -754,69 +810,83 @@ linker(void)
   gp_cofflink_combine_overlay(state.object, 0);
 
   /* combine all sections with the same name */
-  gp_cofflink_merge_sections(state.object, state.byte_addr);
+  gp_cofflink_merge_sections(state.object);
 
   /* create ROM data for initialized data sections */
-  gp_cofflink_make_idata(state.object);
+  gp_cofflink_make_idata(state.object, state.mplink_compatible);
 
-  /* create memory representing target memory */ 
+  /* create memory representing target memory */
   data = i_memory_create();
   program = i_memory_create();
 
   /* allocate memory for absolute sections */
   gp_debug("verifying absolute sections.");
-  gp_cofflink_reloc_abs(program, 
-                        state.byte_addr,
+  gp_cofflink_reloc_abs(program,
+                        state.class->org_to_byte_shift,
                         state.object->sections,
                         STYP_TEXT | STYP_DATA_ROM);
-  gp_cofflink_reloc_abs(data, 
+  gp_cofflink_reloc_abs(data,
                         0,
-                        state.object->sections, 
-                        STYP_DATA | STYP_BSS | STYP_SHARED | 
+                        state.object->sections,
+                        STYP_DATA | STYP_BSS | STYP_SHARED |
                         STYP_OVERLAY | STYP_ACCESS);
 
-  /* FIXME: allocate assigned stacks */ 
+  if (state.mplink_compatible) {
+    /* allocate cinit section to the lowest possible address */
+    gp_section_type *cinit_section;
+
+    cinit_section = gp_coffgen_findsection(state.object,
+                       state.object->sections,
+                       ".cinit");
+    if (cinit_section) {
+      gp_cofflink_reloc_cinit(program,
+                              state.class->org_to_byte_shift,
+                              cinit_section,
+                              state.section.definition);
+    }
+  }
+
+  /* FIXME: allocate assigned stacks */
 
   /* allocate memory for relocatable assigned sections */
   gp_debug("relocating assigned sections.");
-  gp_cofflink_reloc_assigned(program, 
-                             state.byte_addr,
+  gp_cofflink_reloc_assigned(program,
+                             state.class->org_to_byte_shift,
                              state.object->sections,
                              STYP_TEXT | STYP_DATA_ROM,
                              state.section.definition,
                              state.section.logical);
-  gp_cofflink_reloc_assigned(data, 
+  gp_cofflink_reloc_assigned(data,
                              0,
-                             state.object->sections, 
-                             STYP_DATA | STYP_BSS | STYP_SHARED | 
+                             state.object->sections,
+                             STYP_DATA | STYP_BSS | STYP_SHARED |
                              STYP_OVERLAY | STYP_ACCESS,
                              state.section.definition,
                              state.section.logical);
-  
-  /* FIXME: allocate unassigned stacks */ 
+
+  /* FIXME: allocate unassigned stacks */
 
   /* allocate memory for relocatable unassigned sections */
   gp_debug("relocating unassigned sections.");
-  gp_cofflink_reloc_unassigned(program, 
-                               state.byte_addr,
+  gp_cofflink_reloc_unassigned(program,
+                               state.class->org_to_byte_shift,
                                state.object->sections,
                                STYP_TEXT | STYP_DATA_ROM,
                                state.section.definition);
-  gp_cofflink_reloc_unassigned(data, 
+  gp_cofflink_reloc_unassigned(data,
                                0,
-                               state.object->sections, 
-                               STYP_DATA | STYP_BSS | STYP_SHARED | 
+                               state.object->sections,
+                               STYP_DATA | STYP_BSS | STYP_SHARED |
                                STYP_OVERLAY | STYP_ACCESS,
                                state.section.definition);
 
   /* load the table with the relocated addresses */
-  gp_add_cinit_section(state.object, state.byte_addr);
+  gp_add_cinit_section(state.object);
 
-  gp_cofflink_update_table(state.object);
+  gp_cofflink_update_table(state.object, state.class->org_to_byte_shift);
 
   gp_cofflink_fill_pages(state.object,
                          program,
-                         state.byte_addr,
                          state.section.definition);
 
   i_memory_free(data);
@@ -838,14 +908,14 @@ linker(void)
 
   /* convert the executable object into a hex file */
   state.i_memory = gp_cofflink_make_memory(state.object);
-  
+
   /* write hex file */
   writehex(state.basefilename,
            state.i_memory,
            state.hex_format,
            gp_num_errors,
            0,
-           0);
+           state.class->core_size);
 
   /* convert the executable object into a cod file and list file */
   cod_init();
@@ -856,19 +926,15 @@ linker(void)
   make_map();
 
   i_memory_free(state.i_memory);
-  
-  if (gp_num_errors > 0)
-    return EXIT_FAILURE;
-  else
-    return EXIT_SUCCESS;
 
+  return gp_num_errors <= 0;
 }
 
-int 
+int
 main(int argc, char *argv[])
 {
   init();
   process_args(argc, argv);
 
-  return linker();
+  return linker() ? EXIT_SUCCESS : EXIT_FAILURE;
 }
