@@ -29,39 +29,30 @@ Boston, MA 02111-1307, USA.  */
 #define HEAD(L) (L)->value.list.head
 #define TAIL(L) (L)->value.list.tail
 
-struct command {
-  char *name;
-  long int address;
-};
-
-typedef int lkrfunc(char *name, struct pnode *parms);
+#define NELEM(x)  (sizeof(x) / sizeof(x)[0])
 
 void
-script_error(char *messg, char *detail)
+script_error(const char *messg, const char *detail)
 {
-
   gp_num_errors++;
   if (!gp_quiet) {
     if (state.src == NULL) {
-      printf("%s\n", messg); 
+      printf("%s\n", messg);
     } else if (detail == NULL) {
-      printf("%s:%d:Error %s\n", 
-             state.src->name, 
-             state.src->line_number, 
-             messg); 
+      printf("%s:%d:Error %s\n",
+             state.src->name,
+             state.src->line_number,
+             messg);
     } else {
-      printf("%s:%d:Error %s (%s)\n", 
-             state.src->name, 
-             state.src->line_number, 
+      printf("%s:%d:Error %s (%s)\n",
+             state.src->name,
+             state.src->line_number,
              messg,
-             detail);      
+             detail);
     }
   }
-
-  return;
 }
 
-
 static int enforce_simple(struct pnode *p)
 {
   if (p->tag == symbol) {
@@ -72,12 +63,14 @@ static int enforce_simple(struct pnode *p)
   }
 }
 
-static int evaluate(struct pnode *p)
+static long evaluate(struct pnode *p)
 {
 
   switch (p->tag) {
   case constant:
     return p->value.constant;
+  case symbol:
+    return get_script_macro(p->value.symbol);
   default:
     script_error("illegal argument", NULL);
     return 0;
@@ -85,10 +78,10 @@ static int evaluate(struct pnode *p)
 
 }
 
-static int do_files(char *name, struct pnode *parms)
+static int do_files(char *name, enum section_type type, struct pnode *parms)
 {
   struct pnode *p;
-  
+
   for (; parms != NULL; parms = TAIL(parms)) {
     p = HEAD(parms);
     if (enforce_simple(p)) {
@@ -99,7 +92,7 @@ static int do_files(char *name, struct pnode *parms)
   return 0;
 }
 
-static int do_include(char *name, struct pnode *parms)
+static int do_include(char *name, enum section_type type, struct pnode *parms)
 {
   struct pnode *p;
 
@@ -116,7 +109,7 @@ int
 add_path(struct pnode *parms)
 {
   struct pnode *p;
-  
+
   for (; parms != NULL; parms = TAIL(parms)) {
     p = HEAD(parms);
     if (enforce_simple(p)) {
@@ -129,67 +122,119 @@ add_path(struct pnode *parms)
   return 0;
 }
 
-/* FIXME:  These functions were written to allow the user the greatest 
+void add_script_macro(const char *name, long value)
+{
+  struct symbol *sym;
+  long *val;
+  sym = add_symbol(state.script_symbols, name);
+  val = get_symbol_annotation(sym);
+  if (!val) {
+    val = malloc(sizeof value);
+    annotate_symbol(sym, val);
+  }
+  *val = value;
+}
+
+long get_script_macro(const char *name)
+{
+  struct symbol *sym;
+  long *val;
+  sym = get_symbol(state.script_symbols, name);
+  if (!sym)
+    return 0;
+  val = get_symbol_annotation(sym);
+  return *val;
+}
+
+/* FIXME:  These functions were written to allow the user the greatest
    flexibility.  The arguments can appear in any order.  In hind sight
    this may not have been necessary and resulted in overly complicated
    code. */
 
 /* logical section definition */
-static int do_logsec(char *name, struct pnode *parms)
+static int do_logsec(char *name, enum section_type type, struct pnode *parms)
 {
+  enum arg_id_e {
+    id_unknown = 0,
+    id_name = 1,
+    id_ram,
+    id_rom
+  } arg;
+  static struct {
+    enum arg_id_e id;
+    const char *name;
+  }
+  arg_tbl[] = {
+    { id_name, "name" },
+    { id_ram,  "ram"  },
+    { id_rom,  "rom"  }
+  };
   struct pnode *p;
-  gp_boolean found_secname   = false;
-  gp_boolean found_ram       = false;
-  gp_boolean found_rom       = false;
-  char *logical_section_name = NULL;
+  gp_boolean found_secname = false;
+  gp_boolean found_ram = false;
+  gp_boolean found_rom = false;
+  const char *logical_section_name = NULL;
   char *section_name = NULL;
   struct linker_section *section = NULL;
   struct symbol *sym;
-  
+
   /* read the options */
   for (; parms != NULL; parms = TAIL(parms)) {
     p = HEAD(parms);
     if ((p->tag == binop) &&
-	(p->value.binop.op == '=')) {
+        (p->value.binop.op == '=')) {
       if (enforce_simple(p->value.binop.p0)) {
-	char *lhs;
+        const char *lhs = p->value.binop.p0->value.symbol;
+        int i;
 
-	lhs = p->value.binop.p0->value.symbol;
-	if (strcasecmp(lhs, "name") == 0) {
+        arg = id_unknown;
+        for (i = 0; i < NELEM(arg_tbl); ++i) {
+          if (0 == strcasecmp(arg_tbl[i].name, lhs))
+            arg = arg_tbl[i].id;
+        }
+
+        switch (arg) {
+        case id_name:
           if (enforce_simple(p->value.binop.p1)) {
             found_secname = true;
-            logical_section_name = p->value.binop.p1->value.symbol; 
+            logical_section_name = p->value.binop.p1->value.symbol;
           }
-        } else if (strcasecmp(lhs, "ram") == 0) {
+          break;
+
+        case id_ram:
           if (enforce_simple(p->value.binop.p1)) {
             found_ram = true;
             section_name = strdup(p->value.binop.p1->value.symbol);
-	  }
-        } else if (strcasecmp(lhs, "rom") == 0) {
+          }
+          break;
+
+        case id_rom:
           if (enforce_simple(p->value.binop.p1)) {
             found_rom = true;
             section_name = strdup(p->value.binop.p1->value.symbol);
-	  }
-        } else {
-          script_error("illegal argument", lhs);        
-        }       
+          }
+          break;
+
+        default:
+          script_error("illegal argument", lhs);
+        }
       }
     } else {
       if (enforce_simple(p)) {
-        script_error("illegal argument", p->value.symbol);      
+        script_error("illegal argument", p->value.symbol);
       }
     }
   }
 
-  /* process the options */ 
+  /* process the options */
   if (!found_secname) {
-    script_error("missing argument", "name");   
+    script_error("missing argument", "name");
   } else if (found_rom && found_ram) {
     script_error("too many arguments", "ram or rom");
   } else if ((!found_rom) && (!found_ram)) {
     script_error("missing argument", "ram or rom");
   } else {
-    sym = get_symbol(state.section.definition, section_name);      
+    sym = get_symbol(state.section.definition, section_name);
     if (sym == NULL) {
       script_error("undefined section", section_name);
     } else {
@@ -198,10 +243,10 @@ static int do_logsec(char *name, struct pnode *parms)
       if ((found_ram == 1) && (section->type == codepage)) {
         script_error("invalid argument", "ram");
       } else if ((found_rom == 1) && (section->type != codepage)) {
-        script_error("invalid argument", "rom");      
+        script_error("invalid argument", "rom");
       } else {
         sym = add_symbol(state.section.logical, logical_section_name);
-        annotate_symbol(sym, section_name);	       
+        annotate_symbol(sym, section_name);
       }
     }
   }
@@ -210,18 +255,38 @@ static int do_logsec(char *name, struct pnode *parms)
 }
 
 /* general section definition processing */
-static int do_secdef(char *name, struct pnode *parms)
+static int do_secdef(char *name, enum section_type type, struct pnode *parms)
 {
+  enum arg_id_e {
+    id_unknown = 0,
+    id_name = 1,
+    id_start,
+    id_end,
+    id_fill,
+    id_shadow
+  } arg;
+  static struct {
+    enum arg_id_e id;
+    const char *name;
+  }
+  arg_tbl[] = {
+    { id_name,   "name"   },
+    { id_start,  "start"  },
+    { id_end,    "end"    },
+    { id_fill,   "fill"   },
+    { id_shadow, "shadow" }
+  };
   struct pnode *p;
-  gp_boolean found_secname   = false;
-  gp_boolean found_start     = false;
-  gp_boolean found_end       = false;
-  gp_boolean found_fill      = false;
+  gp_boolean found_start = false;
+  gp_boolean found_end = false;
+  gp_boolean found_fill = false;
   gp_boolean found_protected = false;
-  char *section_name = NULL;
-  int  start = 0;
-  int  end = 0;
-  int  fill = 0;
+  const char *section_name = NULL;
+  const char *shadow_sym = NULL;
+  long start = 0;
+  long end = 0;
+  long fill = 0;
+  long shadow_val = 0;
   struct symbol *sym;
   struct linker_section *section_def;
 
@@ -229,73 +294,121 @@ static int do_secdef(char *name, struct pnode *parms)
   for (; parms != NULL; parms = TAIL(parms)) {
     p = HEAD(parms);
     if ((p->tag == binop) &&
-	(p->value.binop.op == '=')) {
+        (p->value.binop.op == '=')) {
       if (enforce_simple(p->value.binop.p0)) {
-	char *lhs;
+        const char *lhs = p->value.binop.p0->value.symbol;
+        int i;
 
-	lhs = p->value.binop.p0->value.symbol;
-	if (strcasecmp(lhs, "name") == 0) {
+        arg = id_unknown;
+        for (i = 0; i < NELEM(arg_tbl); ++i) {
+          if (0 == strcasecmp(arg_tbl[i].name, lhs))
+            arg = arg_tbl[i].id;
+        }
+
+        switch (arg) {
+        case id_name:
           if (enforce_simple(p->value.binop.p1)) {
-            found_secname = true;
-            section_name = p->value.binop.p1->value.symbol; 
+            section_name = p->value.binop.p1->value.symbol;
           }
-        } else if (strcasecmp(lhs, "start") == 0) {
+          break;
+
+        case id_start:
           found_start = true;
           start = evaluate(p->value.binop.p1);
-        } else if (strcasecmp(lhs, "end") == 0) {
+          break;
+
+        case id_end:
           found_end = true;
-          end = evaluate(p->value.binop.p1);             
-        } else if (strcasecmp(lhs, "fill") == 0) {
+          end = evaluate(p->value.binop.p1);
+          break;
+
+        case id_fill:
           found_fill = true;
           fill = evaluate(p->value.binop.p1);
-        } else {
-          script_error("illegal argument", lhs);        
-        }       
-      
+          break;
+
+        case id_shadow:
+          if (enforce_simple(p->value.binop.p1)) {
+            const char *begin;
+            char *end;
+            begin = p->value.binop.p1->value.symbol;
+            if (NULL != (end = strchr(begin, ':'))) {
+              if (end == begin)
+                script_error("bad shadow symbol", lhs);
+              else {
+                char *shsym = (char *)malloc(end - begin + 1);
+                memcpy(shsym, begin, end - begin);
+                shsym[end - begin] = '\0';
+                shadow_sym = shsym;
+                begin = ++end;
+                shadow_val = strtol(begin, &end, 0);
+                if ('\0' != *end)
+                  script_error("bad shadow value", lhs);
+              }
+            } else {
+              script_error("bad shadow argument", lhs);
+            }
+          }
+          break;
+
+        default:
+          script_error("illegal argument", lhs);
+        }
       }
     } else {
       if (enforce_simple(p)) {
-	if (strcasecmp(p->value.symbol, "protected") == 0) {
-          found_protected = true; 	   
-	} else {
+        if (strcasecmp(p->value.symbol, "protected") == 0) {
+          found_protected = true;
+        } else {
           script_error("illegal argument", p->value.symbol);
         }
-      }    
+      }
     }
   }
 
   /* process the options */
-  if (!found_secname) {
+  if (NULL == section_name) {
     script_error("missing argument", "name");
   } else if (!found_start) {
     script_error("missing argument", "start");
   } else if (!found_end) {
     script_error("missing argument", "end");
   } else {
-    sym = get_symbol(state.section.definition, section_name);      
+    sym = get_symbol(state.section.definition, section_name);
     if (sym == NULL) {
       sym = add_symbol(state.section.definition, section_name);
-      section_def = (struct linker_section *)malloc(sizeof(*section_def));
+      section_def = (struct linker_section *)malloc(sizeof(struct linker_section));
+      memset(section_def, 0, sizeof(struct linker_section));
       annotate_symbol(sym, section_def);
 
-      if (strcasecmp(name, "accessbank") == 0) {
+      switch (type) {
+      case accessbank:
         if (state.class != PROC_CLASS_PIC16E) {
-          script_error("accessbank only valid with 18xx devices", name);      
-        }        
-        section_def->type = accessbank;
-      } else if (strcasecmp(name, "codepage") == 0) {
-        section_def->type = codepage;
-      } else if (strcasecmp(name, "databank") == 0) {
-        section_def->type = databank;
-      } else if (strcasecmp(name, "sharebank") == 0) {
-        section_def->type = sharebank;
-      } else {
-        script_error("invalid definition type", name);      
+          script_error("accessbank only valid with 18xx devices", name);
+        }
+        break;
+
+      case codepage:
+        start = gp_processor_org_to_byte(state.class, start);
+        end = gp_processor_org_to_byte(state.class, end + 1) - 1;
+        break;
+
+      case sect_none:
+        script_error("invalid definition type", name);
+        break;
+
+      default:
+        break;
       }
 
+      section_def->type = type;
       section_def->start = start;
       section_def->end = end;
       section_def->protected = found_protected;
+      section_def->shadow_sym = shadow_sym;
+      section_def->shadow_val = shadow_val;
+      section_def->fill = 0;
+      section_def->use_fill = false;
 
       if (section_def->type == codepage) {
         if ((!state.fill_enable) || found_protected) {
@@ -306,23 +419,33 @@ static int do_secdef(char *name, struct pnode *parms)
           section_def->use_fill = true;
         }
       } else if (found_fill == 1) {
-        script_error("illegal argument", "fill");        
+        script_error("illegal argument", "fill");
       }
-    } else if (strcasecmp(name, "sharebank") != 0) {
+    } else if (type != sharebank) {
       script_error("duplicate section definition", section_name);
     }
-
   }
 
   return 0;
 }
 
-static int do_stack(char *name, struct pnode *parms)
+static int do_stack(char *name, enum section_type type, struct pnode *parms)
 {
+  enum arg_id_e {
+    id_unknown = 0,
+    id_size = 1,
+    id_ram
+  } arg;
+  static struct {
+    enum arg_id_e id;
+    const char *name;
+  }
+  arg_tbl[] = {
+    { id_size, "size" },
+    { id_ram,  "ram"  }
+  };
   struct pnode *p;
-
   gp_boolean found_size  = false;
-  gp_boolean found_ram   = false;
   char *ram_name = NULL;
   struct symbol *sym;
 
@@ -339,22 +462,32 @@ static int do_stack(char *name, struct pnode *parms)
   for (; parms != NULL; parms = TAIL(parms)) {
     p = HEAD(parms);
     if ((p->tag == binop) &&
-	(p->value.binop.op == '=')) {
+        (p->value.binop.op == '=')) {
       if (enforce_simple(p->value.binop.p0)) {
-	char *lhs;
+        const char *lhs = p->value.binop.p0->value.symbol;
+        int i;
 
-	lhs = p->value.binop.p0->value.symbol;
-        if (strcasecmp(lhs, "size") == 0) {
+        arg = id_unknown;
+        for (i = 0; i < NELEM(arg_tbl); ++i) {
+          if (0 == strcasecmp(arg_tbl[i].name, lhs))
+            arg = arg_tbl[i].id;
+        }
+
+        switch (arg) {
+        case id_size:
           found_size = true;
           state.stack_size = evaluate(p->value.binop.p1);
-        } else if (strcasecmp(lhs, "ram") == 0) {
+          break;
+
+        case id_ram:
           if (enforce_simple(p->value.binop.p1)) {
-            found_ram = true;
-            ram_name = p->value.binop.p1->value.symbol;   
+            ram_name = strdup(p->value.binop.p1->value.symbol);
           }
-        } else {
-          script_error("illegal argument", lhs);        
-        }       
+          break;
+
+        default:
+          script_error("illegal argument", lhs);
+        }
       }
     } else {
       if (enforce_simple(p)) {
@@ -368,47 +501,38 @@ static int do_stack(char *name, struct pnode *parms)
     script_error("missing argument", "size");
   } else if (ram_name != NULL) {
     sym = add_symbol(state.section.logical, strdup(".stack"));
-    annotate_symbol(sym, ram_name);	       
+    annotate_symbol(sym, ram_name);
   }
 
   return 0;
 }
 
-
-static struct command commands[] = {
-  { "accessbank", (long int)do_secdef     },
-  { "codepage",   (long int)do_secdef     },
-  { "databank",   (long int)do_secdef     },
-  { "files",      (long int)do_files      },
-  { "include",    (long int)do_include    },
-  { "section",    (long int)do_logsec     },
-  { "sharebank",  (long int)do_secdef     },
-  { "stack",	  (long int)do_stack      }
-};
-
-#define NUM_COMMAND  (sizeof(commands) / sizeof(commands[0]))
-
 int execute_command(char *name, struct pnode *parms)
 {
-  int value = 0;
-  int i = 0;
-  gp_boolean found_command = false;
+  static struct {
+    const char *name;
+    enum section_type type;
+    int (*address)(char *name, enum section_type type, struct pnode *parms);
+  } commands[] = {
+    { "accessbank", accessbank, do_secdef  },
+    { "codepage",   codepage,   do_secdef  },
+    { "databank",   databank,   do_secdef  },
+    { "files",      sect_none,  do_files   },
+    { "include",    sect_none,  do_include },
+    { "linearmem",  linearmem,  do_secdef  },
+    { "section",    sect_none,  do_logsec  },
+    { "sharebank",  sharebank,  do_secdef  },
+    { "stack",      sect_none,  do_stack   }
+  };
+  int i;
 
-  while (i < NUM_COMMAND) {
+  for (i = 0; i < NELEM(commands); ++i) {
     if (strcasecmp(name, commands[i].name) == 0) {
-      found_command = true;
-      break;
+      return (*commands[i].address)(name, commands[i].type, parms);
     }
-    i++;
   }
 
-  if (!found_command) {
-    script_error("invalid script command", name);
-  } else {
-    value = (*(lkrfunc*)commands[i].address)(name, parms);
-  }
+  script_error("invalid script command", name);
 
-  return value;
+  return 0;
 }
-
-
