@@ -29,6 +29,7 @@ Boston, MA 02111-1307, USA.  */
 
 #include <stdio.h>
 #include <assert.h>
+#include <sys/stat.h>
 
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
@@ -57,6 +58,17 @@ Boston, MA 02111-1307, USA.  */
 #endif
 #endif
 
+#ifndef HAVE_STRNCASECMP
+#ifdef HAVE_STRNICMP
+#define strncasecmp strnicmp
+#endif
+#endif
+
+#ifndef HAVE_STRNDUP
+/* no system strndup: use the one iplemented in libiberty */
+char *strndup (const char *s, size_t n);
+#endif
+
 #if defined(HAVE_STRING_H)
 #include <string.h>
 #elif defined(HAVE_STRINGS_H)
@@ -66,6 +78,7 @@ Boston, MA 02111-1307, USA.  */
 #include <time.h>
 #include <ctype.h>
 #include <errno.h>
+#include <limits.h>
 
 #ifdef HAVE_DOS_BASED_FILE_SYSTEM
   #define UNIX_PATH_CHAR '/'
@@ -76,6 +89,6 @@ Boston, MA 02111-1307, USA.  */
   #define COPY_CHAR "/"
 #endif
 
-#define GPUTILS_COPYRIGHT_STRING "Copyright (c) 1998-2005 gputils project"
+#define GPUTILS_COPYRIGHT_STRING "Copyright (c) 1998-2014 gputils project"
 
 #endif
