@@ -24,8 +24,8 @@ Boston, MA 02111-1307, USA.  */
 
 /* FIXME: member headers always start on an even-byte boundary. A newline
    character is often used to fill the gap. */
- 
-int 
+
+int
 gp_archive_count_members(gp_archive_type *archive)
 {
   int number = 0;
@@ -33,7 +33,7 @@ gp_archive_count_members(gp_archive_type *archive)
   /* If present, skip the symbol index. */
   if (gp_archive_have_index(archive))
     archive = archive->next;
-  
+
   while (archive != NULL) {
     number++;
     archive = archive->next;
@@ -52,13 +52,13 @@ gp_archive_member_name(gp_archive_type *archive)
 
   sscanf(archive->header.ar_name, "%255s/", name);
   end = strchr(&name[0], '/');
-  if (end != NULL) 
+  if (end != NULL)
     *end = '\0';
 
   return strdup(name);
 }
 
-void 
+void
 gp_archive_list_members(gp_archive_type *archive)
 {
   const char *format = "%-24s  %06i bytes  %s";
@@ -77,7 +77,7 @@ gp_archive_list_members(gp_archive_type *archive)
     sscanf(archive->header.ar_date, "%il", &date);
     sscanf(archive->header.ar_size, "%il", &size);
     end = strchr(&name[0], '/');
-    if (end != NULL) 
+    if (end != NULL)
       *end = '\0';
     time = date;
     printf(format, name, size, ctime(&time));
@@ -101,7 +101,7 @@ gp_archive_find_member(gp_archive_type *archive, char *objectname)
   while (archive != NULL) {
     sscanf(archive->header.ar_name, "%255s/", name);
     end = strrchr(&name[0], '/');
-    if (end != NULL) 
+    if (end != NULL)
       *end = '\0';
     if (strcmp(objectname, name) == 0) {
       found = archive;
@@ -113,12 +113,12 @@ gp_archive_find_member(gp_archive_type *archive, char *objectname)
   return found;
 }
 
-int 
+int
 gp_archive_free_member(gp_archive_type *archive)
 {
 
-  if (archive->file != NULL)
-    free(archive->file);
+  if (archive->data.file != NULL)
+    free(archive->data.file);
   if (archive != NULL)
     free(archive);
 
@@ -126,12 +126,12 @@ gp_archive_free_member(gp_archive_type *archive)
 }
 
 gp_archive_type *
-gp_archive_delete_member(gp_archive_type *archive, 
+gp_archive_delete_member(gp_archive_type *archive,
                          char *objectname)
 {
   gp_archive_type *object = NULL;
   gp_archive_type *list = NULL;
-  
+
   object = gp_archive_find_member(archive, objectname);
   assert(object != NULL);
 
@@ -147,7 +147,7 @@ gp_archive_delete_member(gp_archive_type *archive,
         break;
       }
       list = list->next;
-    }  
+    }
   }
 
   gp_archive_free_member(object);
@@ -156,7 +156,7 @@ gp_archive_delete_member(gp_archive_type *archive,
 }
 
 gp_archive_type *
-gp_archive_add_member(gp_archive_type *archive, 
+gp_archive_add_member(gp_archive_type *archive,
                       char *filename,
                       char *objectname)
 {
@@ -176,19 +176,19 @@ gp_archive_add_member(gp_archive_type *archive,
 
   /* Point the archive member file to the object file.  The object is never
      freed, so this is ok.  It will be cleaned up later */
-  newmember->file = newobject->file;
+  newmember->data = *newobject;
 
   /* fill in the archive header */
   memset(&newmember->header, 0x20, AR_HDR_SIZ); /* fill the header with space */
 
   timer = (int)time(NULL);
-  
+
   snprintf(name, sizeof(name), "%s/", objectname);
   snprintf(date, sizeof(date), "%il", timer);
   snprintf(size, sizeof(size), "%lil", newobject->size);
 
-  /* FIXME:  These functions overwrite the 0x20 that the header is filled 
-     with. */ 
+  /* FIXME:  These functions overwrite the 0x20 that the header is filled
+     with. */
   strncpy(newmember->header.ar_name,
           &name[0],
           sizeof(newmember->header.ar_name));
@@ -203,7 +203,7 @@ gp_archive_add_member(gp_archive_type *archive,
           sizeof(newmember->header.ar_fmag));
 
   oldmember = gp_archive_find_member(archive, objectname);
-  
+
   if (oldmember != NULL) {
     /* the object already exists, replace it */
     archive = gp_archive_delete_member(archive, objectname);
@@ -213,19 +213,19 @@ gp_archive_add_member(gp_archive_type *archive,
     /* the list is empty */
     archive = newmember;
   } else {
-    /* append the new object to the end of the list */  
+    /* append the new object to the end of the list */
     list = archive;
     while (list->next != NULL) {
       list = list->next;
     }
     list->next = newmember;
   }
-  
+
   return archive;
 }
 
-int 
-gp_archive_extract_member(gp_archive_type *archive, 
+int
+gp_archive_extract_member(gp_archive_type *archive,
                           char *objectname)
 {
   gp_archive_type *object = NULL;
@@ -239,9 +239,8 @@ gp_archive_extract_member(gp_archive_type *archive,
   /* if the object doesn't have an extension, add one.  This is done for
      some libs generated with other tools.  It should not be necessary
      for libs generated by gplib. */
-  strncpy(filename, objectname, sizeof(filename));
-  if (strrchr(filename, '.') == NULL)
-    strncat(filename, ".o", sizeof(filename));
+  snprintf(filename, sizeof(filename),
+           "%s%s", objectname, strrchr(filename, '.') == NULL ? ".o" : "");
 
   output_file = fopen(filename, "wb");
   if (output_file == NULL) {
@@ -250,7 +249,7 @@ gp_archive_extract_member(gp_archive_type *archive,
   }
 
   sscanf(object->header.ar_size, "%il", &size);
-  fwrite(object->file, 1, size, output_file);
+  fwrite(object->data.file, 1, size, output_file);
 
   fclose(output_file);
 
@@ -259,8 +258,8 @@ gp_archive_extract_member(gp_archive_type *archive,
   return 0;
 }
 
-int 
-gp_archive_write(gp_archive_type *archive, 
+int
+gp_archive_write(gp_archive_type *archive,
                  char *archivename)
 {
   FILE *output_file;
@@ -279,7 +278,7 @@ gp_archive_write(gp_archive_type *archive,
   while (archive != NULL) {
     fwrite(&archive->header, 1, AR_HDR_SIZ, output_file);
     sscanf(archive->header.ar_size, "%il", &size);
-    fwrite(archive->file, 1, size, output_file);
+    fwrite(archive->data.file, 1, size, output_file);
     archive = archive->next;
   }
 
@@ -291,7 +290,7 @@ gp_archive_write(gp_archive_type *archive,
 /* update the offset numbers for the archive, this is only required
    if a symbol index is created */
 
-void 
+void
 gp_archive_update_offsets(gp_archive_type *archive)
 {
   unsigned int offset = SARMAG;
@@ -301,8 +300,8 @@ gp_archive_update_offsets(gp_archive_type *archive)
     archive->offset = offset;
     sscanf(archive->header.ar_size, "%il", &size);
     offset += AR_HDR_SIZ + size;
-    archive = archive->next;  
-  }  
+    archive = archive->next;
+  }
 
   return;
 }
@@ -327,24 +326,32 @@ gp_archive_read(char *filename)
     exit(1);
   }
 
-  fread(&buffer[0], 1, SARMAG, infile); /* read the magic number */
-  if (strncmp(buffer, ARMAG, SARMAG) != 0) {
+  /* read the magic number */
+  if (SARMAG != fread(&buffer[0], 1, SARMAG, infile) ||
+    0 != strncmp(buffer, ARMAG, SARMAG)) {
     fclose(infile);
     return NULL;
   }
 
   while (feof(infile) == 0) {
+    int n;
+
     /* allocate space for the next archive member */
     new = (gp_archive_type *)malloc(sizeof(*new));
     new->next = NULL;
 
     /* read the archive header */
-    fread(&new->header, AR_HDR_SIZ, 1, infile);
+    n = fread(&new->header, 1, AR_HDR_SIZ, infile);
+    if (AR_HDR_SIZ != n)
+      gp_error("bad archive \"%s\"", filename);
 
     /* read the object file or symbol index into memory */
     sscanf(new->header.ar_size, "%il", &object_size);
-    new->file = (char *)malloc(sizeof(char)*object_size);
-    fread(new->file, sizeof(char), object_size, infile); 
+    new->data.size = object_size;
+    new->data.file = (unsigned char *)malloc(object_size);
+    n = fread(new->data.file, sizeof(char), object_size, infile);
+    if (n != object_size)
+      gp_error("bad archive \"%s\"", filename);
 
     /* insert the new member in the archive list */
     if (archive == NULL) {
@@ -356,15 +363,13 @@ gp_archive_read(char *filename)
       list = new;
     }
 
-    /* Some malformed libs have a couple of extra bytes on the end.  The 
-       while eof test passes, but there are no other members.  Test for 
-       it. */ 
+    /* Some malformed libs have a couple of extra bytes on the end.  The
+       while eof test passes, but there are no other members.  Test for
+       it. */
     fgetpos(infile, &position);
-    fread(&tmpheader, AR_HDR_SIZ, 1, infile);
-    if (feof(infile))
-      break;  
-    fsetpos(infile, &position);    
-
+    if (AR_HDR_SIZ != fread(&tmpheader, 1, AR_HDR_SIZ, infile))
+      break;
+    fsetpos(infile, &position);
   }
 
   gp_archive_update_offsets(archive);
@@ -376,40 +381,40 @@ gp_archive_read(char *filename)
 
 /* Determine if the archive has a symbol index */
 
-int 
+int
 gp_archive_have_index(gp_archive_type *archive)
 {
 
   if ((archive != NULL) && (archive->header.ar_name[0] == '/'))
     return 1;
-    
+
   return 0;
 }
 
-/* Remove the symbol index if one exists */ 
+/* Remove the symbol index if one exists */
 
 gp_archive_type *
 gp_archive_remove_index(gp_archive_type *archive)
 {
   gp_archive_type *member = NULL;
-  
+
   /* If present, remove the symbol index. */
   if (gp_archive_have_index(archive)) {
     member = archive;
     archive = archive->next;
     gp_archive_free_member(member);
   }
-    
+
   return archive;
 }
 
 /* Create a symbol index for the archive.  This is always done to make
-   sure duplicate symbols don't get into the library.  This data can 
+   sure duplicate symbols don't get into the library.  This data can
    and should be stored in the file.  The only reason not to is if you
    need compatibility with other tools.  */
 
-int 
-gp_archive_make_index(gp_archive_type *archive, 
+int
+gp_archive_make_index(gp_archive_type *archive,
                       struct symbol_table *definition)
 {
   gp_object_type *object = NULL;
@@ -419,16 +424,17 @@ gp_archive_make_index(gp_archive_type *archive,
   /* If present, skip the symbol index. */
   if (gp_archive_have_index(archive))
     archive = archive->next;
-        
+
   while (archive != NULL) {
     sscanf(archive->header.ar_name, "%255s/", name);
     end = strchr(&name[0], '/');
-    if (end != NULL) 
+    if (end != NULL)
       *end = '\0';
-    object = gp_convert_file(name, archive->file);
+
+    object = gp_convert_file(name, &archive->data);
     assert(object != NULL);
     gp_link_add_symbols(definition, NULL, object);
-    archive = archive->next;  
+    archive = archive->next;
   }
 
   return 0;
@@ -442,23 +448,23 @@ gp_archive_add_index(struct symbol_table *table,
   gp_archive_type *newmember = NULL;
   gp_archive_type *member = NULL;
   int i;
-  struct symbol **lst, **ps, *s;
-  gp_coffsymbol_type *var;
+  const struct symbol **lst, **ps, *s;
+  const gp_coffsymbol_type *var;
   char size[10];
-  char *ptr;
+  unsigned char *ptr;
   long int indexsize = 0;
-  char *symname;
-  
+  const char *symname;
+
   if ((archive == NULL) || (table == NULL))
     return NULL;
-  
+
   /* sort the index */
   ps = lst = malloc(table->count * sizeof(lst[0]));
   for (i = 0; i < HASH_SIZE; i++)
-    for (s = table->hash_table[i]; s; s = s->next) 
+    for (s = table->hash_table[i]; s; s = s->next)
       *ps++ = s;
   assert(ps == &lst[table->count]);
-  qsort(lst, table->count, sizeof(lst[0]), symbol_compare);  
+  qsort(lst, table->count, sizeof(lst[0]), symbol_compare);
 
   /* determine the symbol index size */
   indexsize = AR_INDEX_NUMBER_SIZ;
@@ -473,12 +479,13 @@ gp_archive_add_index(struct symbol_table *table,
     fprintf(stderr, " error allocating memory\n");
     exit(1);
   }
-  newmember->file = (char *)malloc(sizeof(char)*indexsize);
-  if(!newmember->file) {
+  newmember->data.file = (unsigned char *)malloc(indexsize);
+  if(!newmember->data.file) {
     fprintf(stderr, " error allocating memory\n");
     exit(1);
   }
-  newmember->next = NULL;  
+  newmember->data.size = indexsize;
+  newmember->next = NULL;
 
   /* fill in the archive header */
   memset(&newmember->header, 0x20, AR_HDR_SIZ); /* fill the header with space */
@@ -497,13 +504,13 @@ gp_archive_add_index(struct symbol_table *table,
   archive = newmember;
 
   /* recalculate the file offsets for the symbol table */
-  gp_archive_update_offsets(archive);  
-  
+  gp_archive_update_offsets(archive);
+
   /* write the number of symbols to the member */
-  ptr = archive->file;
+  ptr = archive->data.file;
   gp_putl32(ptr, table->count);
   ptr += 4;
-  
+
   /* write the offsets to the member */
   for (i = 0; i < table->count; i++) {
     var = get_symbol_annotation(lst[i]);
@@ -519,25 +526,25 @@ gp_archive_add_index(struct symbol_table *table,
     memcpy(ptr, symbol_name, symbol_len);
     ptr += symbol_len;
   }
-  
+
   return archive;
 }
 
 /* place the symbol from the archive symbol index in the archive symbol table */
 
-int 
+int
 gp_archive_add_symbol(struct symbol_table *table,
-                      char *name,
+                      const char *name,
                       gp_archive_type *member)
 {
   struct symbol *sym;
 
-  /* Search the for the symbol.  If not found, then add it to 
+  /* Search the for the symbol.  If not found, then add it to
      the global symbol table.  */
   sym = get_symbol(table, name);
   if (sym != NULL) {
     return 1;
-  }  
+  }
   sym = add_symbol(table, name);
   annotate_symbol(sym, member);
 
@@ -545,79 +552,77 @@ gp_archive_add_symbol(struct symbol_table *table,
 }
 
 /* This function reads the symbol index in the archive.  The symbols are
-placed in the archive symbol table.  This table stores the name of the 
-symbol and the archive member that the symbol is defined in. */ 
+placed in the archive symbol table.  This table stores the name of the
+symbol and the archive member that the symbol is defined in. */
 
-void 
+void
 gp_archive_read_index(struct symbol_table *table,
                       gp_archive_type *archive)
 {
   int number = 0;
   int i;
-  char *name;
-  char *offset;
+  const char *name;
+  const unsigned char *offset;
   int offset_value = 0;
   gp_archive_type *list;
-  char *file;
-  
+  const unsigned char *file;
+
   assert(gp_archive_have_index(archive));
-  
-  file = archive->file;
-  
+
+  file = archive->data.file;
+
   /* read the number of symbols */
-  number = gp_getl32(file);  
-  
+  number = gp_getl32(file);
+
   /* set the pointers to the offsets and symbol names */
   offset = &file[AR_INDEX_NUMBER_SIZ];
-  name = offset + (AR_INDEX_OFFSET_SIZ * number);
-  
+  name = (const char*)offset + (AR_INDEX_OFFSET_SIZ * number);
+
   for (i = 0; i < number; i++) {
     /* get the symbol offset from the symbol index */
-    offset_value = gp_getl32(offset); 
+    offset_value = gp_getl32(offset);
 
     /* Locate the object file the symbol is defined in.  The both should
        have the same offset */
     list = archive;
     while (list != NULL) {
       if (list->offset == offset_value)
-        break;    
+        break;
       list = list->next;
     }
-    
+
     assert(list != NULL);
-    
+
     /* add the symbol to the archive symbol table */
     gp_archive_add_symbol(table, name, list);
 
     /* increment the pointers to the next symbol */
     name += strlen(name) + 1;
     offset += AR_INDEX_OFFSET_SIZ;
-  }  
-
-  return;
+  }
 }
 
 /* print the archive symbol table */
 
-void 
+void
 gp_archive_print_table(struct symbol_table *table)
 {
-  struct symbol **lst, **ps, *s;
+  const struct symbol **lst, **ps, *s;
   int i;
   const char *format = "%-32s %s\n";
-  gp_archive_type *member;
+  const gp_archive_type *member;
   char name[256];
   char *end;
-  
+
   assert(table != NULL);
-  
+
   /* sort the index */
   ps = lst = malloc(table->count * sizeof(lst[0]));
   for (i = 0; i < HASH_SIZE; i++)
-    for (s = table->hash_table[i]; s; s = s->next) 
+    for (s = table->hash_table[i]; s; s = s->next)
       *ps++ = s;
   assert(ps == &lst[table->count]);
-  qsort(lst, table->count, sizeof(lst[0]), symbol_compare);    
+  qsort(lst, table->count, sizeof(lst[0]), symbol_compare);
 
   for (i = 0; i < table->count; i++) {
     /* determine the archive member the symbol is defined in */
@@ -626,11 +631,9 @@ gp_archive_print_table(struct symbol_table *table)
     /* determine the archive member name */
     sscanf(member->header.ar_name, "%255s/", name);
     end = strchr(&name[0], '/');
-    if (end != NULL) 
+    if (end != NULL)
       *end = '\0';
     /* print it */
     printf(format, get_symbol_name(lst[i]), name);
   }
-
-  return;
 }
