@@ -37,7 +37,7 @@ struct filehdr
 {
   unsigned short  f_magic;   /* magic number */
   unsigned short  f_nscns;   /* number of sections */
-  unsigned long   f_timdat;            /* time and date stamp */
+  unsigned long   f_timdat;  /* time and date stamp */
   unsigned long   f_symptr;  /* file ptr to symtab */
   unsigned long   f_nsyms;   /* # symtab entries */
   unsigned short  f_opthdr;  /* sizeof(opt hdr) */
@@ -49,24 +49,28 @@ struct filehdr
 #define FILE_HDR_SIZ_v2 20
 
 /* relocation info has been stripped */
-#define F_RELFLG    0x01
+#define F_RELFLG      0x0001
 /* file is executable - has no unresolved external symbols */
-#define F_EXEC      0x02
+#define F_EXEC        0x0002
 /* line numbers have been stripped */
-#define F_LNNO      0x04
+#define F_LNNO        0x0004
+/* the MPASM assembler object file is from absolute (as opposed to relocatable) assembly code */
+#define F_ABSOLUTE    0x0010
 /* local symbols have been stripped */
-#define L_SYMS      0x80
+#define L_SYMS        0x0080
+/* the COFF file produced utilizing the Extended mode */
+#define F_EXTENDED18  0x4000
 /* processor independent file for a core */
-#define F_GENERIC   0x8000
+#define F_GENERIC     0x8000
 
 /* optional header format */
 struct opthdr
 {
-  unsigned short  opt_magic;                   
-  unsigned long   vstamp;         /* version of the compiler assembler */             
-  unsigned long   proc_type;                   
-  unsigned long   rom_width_bits;                   
-  unsigned long   ram_width_bits;           
+  unsigned short  opt_magic;
+  unsigned long   vstamp;         /* version of the compiler assembler */
+  unsigned long   proc_type;
+  unsigned long   rom_width_bits;
+  unsigned long   ram_width_bits;
 };
 
 #define OPT_HDR_SIZ_v1 16
@@ -76,7 +80,7 @@ struct opthdr
 struct scnhdr
 {
   union {
-    char            name[8];   /* section name if less then 8 characters */
+    char          name[8];   /* section name if less then 8 characters */
     struct {
       unsigned long s_zeros;   /* first four characters are 0 */
       unsigned long s_offset;  /* pointer to the string table */
@@ -108,7 +112,7 @@ struct scnhdr
 #define STYP_ABS      0x1000
 /* Section is shared across banks */
 #define STYP_SHARED   0x2000
-/* Section is overlaid with other sections of the same name from different 
+/* Section is overlaid with other sections of the same name from different
    objects modules */
 #define STYP_OVERLAY  0x4000
 /* Section is available using access bit */
@@ -123,9 +127,9 @@ struct scnhdr
 /* relocation entry */
 struct reloc
 {
-  unsigned long            r_vaddr;     /* entry relative virtual address */
-  unsigned long            r_symndx;    /* index into symbol table */
-  short           r_offset;    /* offset to be added to address of symbol 
+  unsigned long   r_vaddr;     /* entry relative virtual address */
+  unsigned long   r_symndx;    /* index into symbol table */
+  short           r_offset;    /* offset to be added to address of symbol
                                   'r_symndx' */
   unsigned short  r_type;      /* relocation type */
 };
@@ -140,20 +144,20 @@ struct reloc
 #define     RELOCT_HIGH     3
 /* relocation for the low order 8 bits of an address */
 #define     RELOCT_LOW      4
-/* relocation for the 5 bits of address for the P operand of a 17cxx MOVFP 
+/* relocation for the 5 bits of address for the P operand of a 17cxx MOVFP
    or MOVPF instruction */
 #define     RELOCT_P        5
-/* relocation to generate the appropriate instruction to bank switch for a 
+/* relocation to generate the appropriate instruction to bank switch for a
    symbol */
 #define     RELOCT_BANKSEL  6
-/* relocation to generate the appropriate instruction to page switch for a 
+/* relocation to generate the appropriate instruction to page switch for a
    symbol */
 #define     RELOCT_PAGESEL  7
 /* FIXME */
 #define     RELOCT_ALL      8
 /* FIXME */
 #define     RELOCT_IBANKSEL 9
-/* relocation for the 8 bits of address for the F operand of a 17cxx MOVFP or 
+/* relocation for the 8 bits of address for the F operand of a 17cxx MOVFP or
    MOVPF instruction */
 #define     RELOCT_F        10
 /* FIXME */
@@ -199,12 +203,14 @@ struct reloc
 /* relocation for the address of the end of a section on LFSR */
 #define     RELOCT_SCNEND_LFSR1 31
 #define     RELOCT_SCNEND_LFSR2 32
-#define	    RELOCT_TRIS_4BIT    33
+#define     RELOCT_TRIS_3BIT    33
+/* relocation for selecting the correct page using pic14 enhanced MOVLP instruction */
+#define     RELOCT_PAGESEL_MOVLP   34
 
 /* linenumber entry */
 struct lineno
 {
-  unsigned long  l_srcndx;     /* symbol table index of associated source 
+  unsigned long  l_srcndx;     /* symbol table index of associated source
                                   file */
   unsigned short l_lnno;       /* line number */
   unsigned long  l_paddr;      /* address of code for this lineno */
@@ -246,9 +252,9 @@ struct syment
 
 /* Symbol types
 
-  Type in an unsigned short (16 bits).  The lowest nibble contains the basic
-  type.  The next higher three bits contain the derived symbol type.  The
-  rest of the bits are unused. */
+  Type in an unsigned short (16 bits).  The lowest four or five bits contains
+  the basic type.  The next higher three bits contain the derived symbol type.
+  The rest of the bits are unused. */
 
 /* Basic symbol types  */
 #define T_NULL      0   /* null */
@@ -267,16 +273,17 @@ struct syment
 #define T_USHORT    13  /* unsigned short */
 #define T_UINT      14  /* unsigned integer */
 #define T_ULONG     15  /* unsigned long */
+/* Basic symbol types for new format only */
 #define T_LNGDBL    16  /* long double floating point */
 #define T_SLONG     17  /* short long */
 #define T_USLONG    18  /* unsigned short long */
 
 /* Derived types */
-#define DT_NON      0  /* no derived type */         
+#define DT_NON      0  /* no derived type */
 #define DT_PTR      1  /* pointer */
 #define DT_FCN      2  /* function */
 #define DT_ARY      3  /* array */
-#define DT_ROMPTR   4  
+#define DT_ROMPTR   4
 #define DT_FARROMPTR 5
 
 /* Storage classes */
@@ -317,7 +324,7 @@ struct syment
 struct aux_file
 {
   unsigned long  x_offset;  /* String table offset for filename */
-  unsigned long  x_incline; /* Line number at which this file was included, 
+  unsigned long  x_incline; /* Line number at which this file was included,
                               0->not included */
   unsigned char  x_flags;
   char _unused[11];
@@ -338,7 +345,7 @@ struct aux_tag
   char _unused[6];
   unsigned short x_size;     /* Size of struct/union/enum */
   char _unused2[4];
-  unsigned long x_endndx;    /* Symbol index of next entry beyond this 
+  unsigned long x_endndx;    /* Symbol index of next entry beyond this
                                 struct/union/enum */
   char _unused3[4];
 };
@@ -358,7 +365,7 @@ struct aux_fcn
   unsigned long x_tagndx;   /* Unused??  Tag Index */
   unsigned long x_size;     /* Unused??  Size of function in bits */
   unsigned long x_lnnoptr;  /* File pointer to line numbers for this function */
-  unsigned long x_endndx;   /* Symbol Index of next entry beyond this 
+  unsigned long x_endndx;   /* Symbol Index of next entry beyond this
                                function */
   short x_actscnum; /* size of static activation record to allocate */
   char _unused[2];
@@ -378,7 +385,7 @@ struct aux_arr
 struct aux_eobf
 {
   char _unused[4];
-  unsigned short x_lnno;   /* C Source line number of the end, relative to 
+  unsigned short x_lnno;   /* C Source line number of the end, relative to
                               start of block/func */
   char _unused2[14];
 };
@@ -387,7 +394,7 @@ struct aux_eobf
 struct aux_bobf
 {
   char _unused[4];
-  unsigned short x_lnno;   /* C Source line number of the beginning, relative 
+  unsigned short x_lnno;   /* C Source line number of the beginning, relative
                               to start enclosing scope */
   char _unused2[6];
   unsigned long x_endndx;  /* Symbol Index of next entry past this block/func */
@@ -410,11 +417,11 @@ struct aux_field {
 };
 
 struct aux_fcn_calls {
-  unsigned long x_calleendx;
-  unsigned long x_is_interrupt;
+  unsigned long x_calleendx;    /* Symbol table entry of callee - 1 */
+  unsigned long x_is_interrupt; /* 0 not, 1 low, 2 high */
   char _unused[12];
 };
-  
+
 
 
 /* Auxiliary entries */
@@ -431,14 +438,15 @@ struct aux_fcn_calls {
 #define AUX_VAR    9  /* variable */
 #define AUX_DIRECT 10 /* direct message */
 #define AUX_IDENT  11 /* ident */
+#define AUX_FCN_CALLS 12  /* function called by this function */
 
 /* These definitions are for the COFF as stored in memory. */
 
 /* relocation linked list */
 
-typedef struct gp_reloc_type 
+typedef struct gp_reloc_type
 {
-  /* entry relative address */
+  /* entry relative address in bytes */
   long address;
 
   /* symbol */
@@ -458,23 +466,23 @@ typedef struct gp_reloc_type
 
 /* line number linked list */
 
-typedef struct gp_linenum_type 
+typedef struct gp_linenum_type
 {
   /* source file symbol */
   struct gp_symbol_type *symbol;
 
   /* line number */
-  unsigned short line_number; 
+  unsigned short line_number;
 
-  /* address of code for this line number */
-  unsigned long address;     
+  /* byte address for this line number */
+  unsigned int address;
 
   struct gp_linenum_type *next;
 } gp_linenum_type;
 
 /* auxiliary symbol linked list */
 
-typedef struct gp_aux_type 
+typedef struct gp_aux_type
 {
   /* auxiliary symbol type */
   long type;
@@ -487,7 +495,8 @@ typedef struct gp_aux_type
     } _aux_direct;
     struct {
       char *filename;
-      unsigned long line_number; 
+      unsigned long line_number;
+      unsigned char flags;
     } _aux_file;
     struct {
       char *string;
@@ -497,31 +506,38 @@ typedef struct gp_aux_type
       unsigned short nreloc;
       unsigned short nlineno;
     } _aux_scn;
+    struct {
+      struct gp_symbol_type *callee; /* NULL for call through pointer */
+      unsigned long is_interrupt; /* 0 not, 1 low, 2 high */
+    } _aux_fcn_calls;
     char data[SYMBOL_SIZE_v2];
   } _aux_symbol;
-  
+
   struct gp_aux_type *next;
 } gp_aux_type;
 
 /* symbol linked list */
 
-typedef struct gp_symbol_type 
+typedef struct gp_symbol_type
 {
   /* symbol name */
   char *name;
 
-  /* symbol value */ 
+  /* symbol value */
   long value;
 
-  /* section number, only for used for determining symbol type: 
+  /* section number, only for used for determining symbol type:
      N_DEBUG = -2, N_ABS = -1, N_UNDEF = 0, or N_SCNUM = 1 if defined */
   short section_number;
-  
-  /* defining section */ 
+
+  /* defining section */
   struct gp_section_type *section;
 
   /* type */
-  unsigned long type;
+  unsigned char type;
+
+  /* derived type */
+  unsigned long derived_type;
 
   /* storage class */
   signed char class;
@@ -532,7 +548,7 @@ typedef struct gp_symbol_type
   /* auxiliary symbols */
   struct gp_aux_type *aux_list;
 
-  /* symbol number, only valid when writing coff or cod file */  
+  /* symbol number, only valid when writing coff or cod file */
   unsigned long number;
 
   struct gp_symbol_type *next;
@@ -540,7 +556,7 @@ typedef struct gp_symbol_type
 
 /* section linked list */
 
-typedef struct gp_section_type 
+typedef struct gp_section_type
 {
   /* section name */
   char *name;
@@ -550,13 +566,19 @@ typedef struct gp_section_type
 
   /* flags */
   unsigned long flags;
-  
-  /* section address */
-  unsigned long address;
-  
+
+  /* section physical address in bytes */
+  unsigned int address;
+
+  /* section virtual address in bytes */
+  unsigned int virtual_address;
+
+  /* section shadow address in bytes */
+  unsigned int shadow_address;
+
   /* section size in bytes */
   unsigned long size;
-  
+
   /* memory linked list */
   MemBlock *data;
 
@@ -578,37 +600,25 @@ typedef struct gp_section_type
   /* tail of line numbers */
   gp_linenum_type *line_numbers_tail;
 
-  /* this section required for symbol resolution, only valid when linking */  
+  /* this section required for symbol resolution, only valid when linking */
   gp_boolean is_used;
 
-  /* section number, only valid when writing coff file */  
+  /* section number, only valid when writing coff file */
   unsigned long number;
 
-  /* data pointer, only valid when writing coff file */  
+  /* data pointer, only valid when writing coff file */
   unsigned long data_ptr;
 
-  /* relocations pointer, only valid when writing coff file */  
+  /* relocations pointer, only valid when writing coff file */
   unsigned long reloc_ptr;
 
-  /* linenumber pointer, only valid when writing coff file */  
+  /* linenumber pointer, only valid when writing coff file */
   unsigned long lineno_ptr;
 
-  /*
-   * when the section is STYP_BPACK, this contains the flag as to whether the
-   * next byte would occupy part of the existing org address or start a new one
-   */
-  gp_boolean have_pack_byte;
-
-  /*
-   * when the section is STYP_BPACK, this will be true when we've just emitted
-   * a packed byte. (for synchronizing the listing)
-   */
-  gp_boolean emitted_pack_byte;
-
   struct gp_section_type *next;
 } gp_section_type;
 
-typedef struct gp_object_type 
+typedef struct gp_object_type
 {
   /* object filename */
   char *filename;
@@ -624,16 +634,16 @@ typedef struct gp_object_type
 
   /* processor */
   pic_processor_t processor;
-  
+
   /* processor class */
-  enum proc_class class;
-  
+  proc_class_t class;
+
   /* time object was created */
   time_t time;
-  
+
   /* flags */
   unsigned short flags;
-  
+
   /* number of sections */
   long num_sections;
 
@@ -645,14 +655,14 @@ typedef struct gp_object_type
 
   /* number of symbols */
   long num_symbols;
-  
+
   /* head of symbols */
   gp_symbol_type *symbols;
 
   /* tail of symbols */
   gp_symbol_type *symbols_tail;
 
-  /* symbol table pointer, only valid when writing coff file */  
+  /* symbol table pointer, only valid when writing coff file */
   unsigned long symbol_ptr;
 
   /* next object in the linked list */
