@@ -25,17 +25,15 @@ Boston, MA 02111-1307, USA.  */
 /* Base the hash func on the 1st, 2nd, 3rd and last characters of the
    string, and its length. */
 
-static int hashfunc(struct symbol_table *t, char *s)
+static int hashfunc_len(struct symbol_table *t, const char *s, size_t len)
 {
   union {
     char b[4];
     unsigned long ul;
   } change;
-  int len;
 
   change.ul = 0;
 
-  len = strlen(s);
   change.b[0] = s[0];
   if (len > 1) {
     change.b[1] = s[1];
@@ -52,16 +50,35 @@ static int hashfunc(struct symbol_table *t, char *s)
   return (change.ul % HASH_SIZE);
 }
 
+static int hashfunc(struct symbol_table *t, const char *s)
+{
+  return hashfunc_len(t, s, strlen(s));
+}
+
+static int compare_len_case(const char *s1, const char *s2, size_t n)
+{
+  return (n == strlen(s2)) ? strncasecmp(s1, s2, n) : -1;
+}
+
+static int compare_len(const char *s1, const char *s2, size_t n)
+{
+  return (n == strlen(s2)) ? strncmp(s1, s2, n) : -1;
+}
+
 struct symbol_table *push_symbol_table(struct symbol_table * table,
-				       gp_boolean case_insensitive)
+                                       gp_boolean case_insensitive)
 {
   struct symbol_table *new = calloc(sizeof(*new), 1);
 
   new->case_insensitive = case_insensitive;
-  if (case_insensitive)
+  if (case_insensitive) {
     new->compare = strcasecmp;
-  else
+    new->compare_len = compare_len_case;
+  }
+  else {
     new->compare = strcmp;
+    new->compare_len = compare_len;
+  }
   new->prev = table;
   return new;
 }
@@ -75,7 +92,7 @@ struct symbol_table *pop_symbol_table(struct symbol_table *table)
   return prev;
 }
 
-struct symbol *add_symbol(struct symbol_table *table, char *name)
+struct symbol *add_symbol(struct symbol_table *table, const char *name)
 {
   struct symbol *r;
   int index;
@@ -97,16 +114,16 @@ struct symbol *add_symbol(struct symbol_table *table, char *name)
     table->hash_table[index] = r;
     table->count++;
   }
-  
+
   return r;
 }
 
 /* FIXME: remove_symbol does not search all of the symbol tables in the stack.
 Maybe this is ok, but it seems wrong. */
 
-int remove_symbol(struct symbol_table *table, char *name)
+int remove_symbol(struct symbol_table *table, const char *name)
 {
-  struct symbol *r = NULL; 
+  struct symbol *r = NULL;
   struct symbol *last = NULL;
   int index;
   int found_symbol = 0;
@@ -141,22 +158,47 @@ int remove_symbol(struct symbol_table *table, char *name)
   return found_symbol;
 }
 
-struct symbol *get_symbol(struct symbol_table *table, char *name)
+struct symbol *get_symbol(struct symbol_table *table, const char *name)
 {
   struct symbol *r = NULL;
 
   assert(name != NULL);
 
-  if (table != NULL) {
+  while (table != NULL) {
     int index = hashfunc(table, name);
 
     r = table->hash_table[index];
     while (r && ((*table->compare)(name, r->name) != 0))
       r = r->next;
 
+    if (NULL != r)
+      break;
+
     /* If r is still NULL, we didn't match.  Try the prev table on the stack */
-    if (r == NULL)
-      r = get_symbol(table->prev, name);
+    table = table->prev;
+  }
+
+  return r;
+}
+
+struct symbol *get_symbol_len(struct symbol_table *table, const char *name, size_t len)
+{
+  struct symbol *r = NULL;
+
+  assert(name != NULL);
+
+  while (table != NULL) {
+    int index = hashfunc_len(table, name, len);
+
+    r = table->hash_table[index];
+    while (r && ((*table->compare_len)(name, r->name, len) != 0))
+      r = r->next;
+
+    if (NULL != r)
+      break;
+
+    /* If r is still NULL, we didn't match.  Try the prev table on the stack */
+    table = table->prev;
   }
 
   return r;
@@ -167,12 +209,12 @@ void annotate_symbol(struct symbol *s, void *a)
   s->annotation = a;
 }
 
-char *get_symbol_name(struct symbol *s)
+const char *get_symbol_name(const struct symbol *s)
 {
   return s->name;
 }
 
-void *get_symbol_annotation(struct symbol *s)
+void *get_symbol_annotation(const struct symbol *s)
 {
   return s->annotation;
 }
